@@ -44,9 +44,12 @@
 #include "xmalloc.h"
 #include "oauth.h"
 
-#ifndef WIN // getpid() on POSIX systems
+#ifndef WIN32 // getpid() on POSIX systems
 #include <sys/types.h>
 #include <unistd.h>
+#else
+#define snprintf _snprintf
+#define strncasecmp strnicmp
 #endif
 
 /**
@@ -218,8 +221,8 @@ char *oauth_url_escape(const char *string) {
       newlen += 2; /* this'll become a %XX */
       if(newlen > alloc) {
         alloc *= 2;
-				testing_ptr = (char*) xrealloc(ns, alloc);
-				ns = testing_ptr;
+        testing_ptr = (char*) xrealloc(ns, alloc);
+        ns = testing_ptr;
       }
       snprintf(&ns[strindex], 4, "%%%02X", in);
       strindex+=3;
@@ -249,7 +252,7 @@ char *oauth_url_unescape(const char *string, size_t *olen) {
   unsigned char in;
   long hex;
 
-	if (!string) return NULL;
+  if (!string) return NULL;
   alloc = strlen(string)+1;
   ns = (char*) xmalloc(alloc);
 
@@ -339,21 +342,21 @@ char *oauth_sign_rsa_sha1 (const char *m, const char *k) {
 
   if (pkey == NULL) {
   //fprintf(stderr, "liboauth/ssl: can not read private key\n");
-	  return xstrdup("liboauth/ssl: can not read private key");
+    return xstrdup("liboauth/ssl: can not read private key");
   }
 
   len = EVP_PKEY_size(pkey);
-  sig = xmalloc((len+1)*sizeof(char));
+  sig = (unsigned char*)xmalloc((len+1)*sizeof(char));
 
   EVP_SignInit(&md_ctx, EVP_sha1());
   EVP_SignUpdate(&md_ctx, m, strlen(m));
   if (EVP_SignFinal (&md_ctx, sig, &len, pkey)) {
-	  char *tmp;
+    char *tmp;
     sig[len] = '\0';
-		tmp = oauth_encode_base64(len,sig);
-		OPENSSL_free(sig);
-	  EVP_PKEY_free(pkey);
-		return tmp;
+    tmp = oauth_encode_base64(len,sig);
+    OPENSSL_free(sig);
+    EVP_PKEY_free(pkey);
+    return tmp;
   }
   return xstrdup("liboauth/ssl: rsa-sha1 signing failed");
 }
@@ -374,33 +377,33 @@ int oauth_verify_rsa_sha1 (const char *m, const char *c, const char *s) {
   EVP_PKEY *pkey;
   BIO *in;
   X509 *cert = NULL;
-	unsigned char *b64d;
+  unsigned char *b64d;
   int slen, err;
 
   in = BIO_new_mem_buf((unsigned char*)c, strlen(c));
   cert = PEM_read_bio_X509(in, NULL, 0, NULL);
-	if (cert)  {
+  if (cert)  {
     pkey = (EVP_PKEY *) X509_get_pubkey(cert); 
-		X509_free(cert);
-	} else {
+    X509_free(cert);
+  } else {
     pkey = PEM_read_bio_PUBKEY(in, NULL, 0, NULL);
-	}
+  }
   BIO_free(in);
   if (pkey == NULL) {
   //fprintf(stderr, "could not read cert/pubkey.\n");
-	  return -2;
+    return -2;
   }
 
   b64d= (unsigned char*) xmalloc(sizeof(char)*strlen(s));
   slen = oauth_decode_base64(b64d, s);
 
-	EVP_VerifyInit(&md_ctx, EVP_sha1());
-	EVP_VerifyUpdate(&md_ctx, m, strlen(m));
-	err = EVP_VerifyFinal(&md_ctx, b64d, slen, pkey);
-	EVP_MD_CTX_cleanup(&md_ctx);
-	EVP_PKEY_free(pkey);
-	free(b64d);
-	return (err);
+  EVP_VerifyInit(&md_ctx, EVP_sha1());
+  EVP_VerifyUpdate(&md_ctx, m, strlen(m));
+  err = EVP_VerifyFinal(&md_ctx, b64d, slen, pkey);
+  EVP_MD_CTX_cleanup(&md_ctx);
+  EVP_PKEY_free(pkey);
+  free(b64d);
+  return (err);
 }
 
 /**
@@ -410,7 +413,7 @@ int oauth_verify_rsa_sha1 (const char *m, const char *c, const char *s) {
  * all arguments thereafter must be of type (char *) 
  *
  * @param len the number of arguments to follow this parameter
- * @param ... string to escape and added
+ * @param ... string to escape and added (may be NULL)
  *
  * @return pointer to memory holding the concatenated 
  * strings - needs to be free(d) by the caller. or NULL
@@ -478,32 +481,32 @@ int oauth_split_post_paramters(const char *url, char ***argv, short qesc) {
     if(!strncasecmp("oauth_signature=",token,16)) continue;
     (*argv)=(char**) xrealloc(*argv,sizeof(char*)*(argc+1));
     while (!(qesc&2) && (tmp=strchr(token,'\001'))) *tmp='&';
-		(*argv)[argc]=oauth_url_unescape(token, NULL);
-	  if (argc==0 && strstr(token, ":/")) {
-			// HTTP does not allow empty absolute paths, so the URL 
-			// 'http://example.com' is equivalent to 'http://example.com/' and should
-			// be treated as such for the purposes of OAuth signing (rfc2616, section 3.2.1)
-			// see http://groups.google.com/group/oauth/browse_thread/thread/c44b6f061bfd98c?hl=en
-			char *slash=strstr(token, ":/");
-			while (slash && *(++slash) == '/')  ; // skip slashes eg /xxx:[\/]*/
+    (*argv)[argc]=oauth_url_unescape(token, NULL);
+    if (argc==0 && strstr(token, ":/")) {
+      // HTTP does not allow empty absolute paths, so the URL 
+      // 'http://example.com' is equivalent to 'http://example.com/' and should
+      // be treated as such for the purposes of OAuth signing (rfc2616, section 3.2.1)
+      // see http://groups.google.com/group/oauth/browse_thread/thread/c44b6f061bfd98c?hl=en
+      char *slash=strstr(token, ":/");
+      while (slash && *(++slash) == '/')  ; // skip slashes eg /xxx:[\/]*/
 #if 0
-			// skip possibly unescaped slashes in the userinfo - they're not allowed by RFC2396 but have been seen.
-			// the hostname/IP may only contain alphanumeric characters - so we're safe there.
-			if (slash && strchr(slash,'@')) slash=strchr(slash,'@'); 
+      // skip possibly unescaped slashes in the userinfo - they're not allowed by RFC2396 but have been seen.
+      // the hostname/IP may only contain alphanumeric characters - so we're safe there.
+      if (slash && strchr(slash,'@')) slash=strchr(slash,'@'); 
 #endif
-			if (slash && !strchr(slash,'/')) {
+      if (slash && !strchr(slash,'/')) {
 #ifdef DEBUG_OAUTH
-			  fprintf(stderr, "\nliboauth: added trailing slash to URL: '%s'\n\n", token);
+        fprintf(stderr, "\nliboauth: added trailing slash to URL: '%s'\n\n", token);
 #endif
-				free((*argv)[argc]);
-				(*argv)[argc]= (char*) xmalloc(sizeof(char)*(2+strlen(token))); 
-				strcpy((*argv)[argc],token);
-				strcat((*argv)[argc],"/");
-		  }
-		}
-	  if (argc==0 && (tmp=strstr((*argv)[argc],":80/"))) {
-			  memmove(tmp, tmp+3, strlen(tmp+2));
-		}
+        free((*argv)[argc]);
+        (*argv)[argc]= (char*) xmalloc(sizeof(char)*(2+strlen(token))); 
+        strcpy((*argv)[argc],token);
+        strcat((*argv)[argc],"/");
+      }
+    }
+    if (argc==0 && (tmp=strstr((*argv)[argc],":80/"))) {
+        memmove(tmp, tmp+3, strlen(tmp+2));
+    }
     tmp=NULL;
     argc++;
   }
@@ -526,7 +529,7 @@ int oauth_split_url_parameters(const char *url, char ***argv) {
  *
  */
 char *oauth_serialize_url (int argc, int start, char **argv) {
-	return oauth_serialize_url_sep( argc, start, argv, "&");
+  return oauth_serialize_url_sep( argc, start, argv, "&");
 }
 
 /**
@@ -541,18 +544,18 @@ char *oauth_serialize_url (int argc, int start, char **argv) {
 char *oauth_serialize_url_sep (int argc, int start, char **argv, char *sep) {
   char  *tmp, *t1;
   int i;
-  int	first=0;
-	int seplen=strlen(sep);
+  int  first=0;
+  int seplen=strlen(sep);
   char *query = (char*) xmalloc(sizeof(char)); 
   *query='\0';
   for(i=start; i< argc; i++) {
     int len = 0;
     if (query) len+=strlen(query);
 
-		if (i==start && i==0 && strstr(argv[i], ":/")) {
+    if (i==start && i==0 && strstr(argv[i], ":/")) {
       tmp=xstrdup(argv[i]);
       len+=strlen(tmp);
-		} else if(!(t1=strchr(argv[i], '='))) {
+    } else if(!(t1=strchr(argv[i], '='))) {
     // see http://oauth.net/core/1.0/#anchor14
     // escape parameter names and arguments but not the '='
       tmp=xstrdup(argv[i]);
@@ -570,15 +573,15 @@ char *oauth_serialize_url_sep (int argc, int start, char **argv, char *sep) {
       free(t1);
       len+=strlen(tmp);
     }
-		len+=seplen+1;
+    len+=seplen+1;
     query=(char*) xrealloc(query,len*sizeof(char));
     strcat(query, ((i==start||first)?"":sep));
-		first=0;
+    first=0;
     strcat(query, tmp);
-		if (i==start && i==0 && strstr(tmp, ":/")) {
-			strcat(query, "?");
-			first=1;
-		}
+    if (i==start && i==0 && strstr(tmp, ":/")) {
+      strcat(query, "?");
+      first=1;
+    }
     free(tmp);
   }
   return (query);
@@ -610,15 +613,15 @@ char *oauth_gen_nonce() {
   char *nc;
   static int rndinit = 1;
   const char *chars = "abcdefghijklmnopqrstuvwxyz"
-  	"ABCDEFGHIJKLMNOPQRSTUVWXYZ" "0123456789_";
+    "ABCDEFGHIJKLMNOPQRSTUVWXYZ" "0123456789_";
   unsigned int max = strlen( chars );
   int i, len;
 
   if(rndinit) {srand(time(NULL) 
-#ifndef WIN // quick windows check.
-  	* getpid()
+#ifndef WIN32 // quick windows check.
+    * getpid()
 #endif
-	); rndinit=0;} // seed random number generator - FIXME: we can do better ;)
+  ); rndinit=0;} // seed random number generator - FIXME: we can do better ;)
 
   len=15+floor(rand()*16.0/(double)RAND_MAX);
   nc = (char*) xmalloc((len+1)*sizeof(char));
@@ -678,11 +681,11 @@ int oauth_cmpstringp(const void *p1, const void *p2) {
  * @return FALSE (0) if array does not contain a paramater with given key, TRUE (1) otherwise.
  */
 int oauth_param_exists(char **argv, int argc, char *key) {
-	int i;
-	size_t l= strlen(key);
-	for (i=0;i<argc;i++)
-		if (strlen(argv[i])>l && !strncmp(argv[i],key,l) && argv[i][l] == '=') return 1;
-	return 0;
+  int i;
+  size_t l= strlen(key);
+  for (i=0;i<argc;i++)
+    if (strlen(argv[i])>l && !strncmp(argv[i],key,l) && argv[i][l] == '=') return 1;
+  return 0;
 }
 
 /**
@@ -708,72 +711,44 @@ void oauth_add_protocol(int *argcp, char ***argvp,
   char oarg[1024];
 
   // add oAuth specific arguments
-	if (!oauth_param_exists(*argvp,*argcp,"oauth_nonce")) {
-		char *tmp;
-		snprintf(oarg, 1024, "oauth_nonce=%s", (tmp=oauth_gen_nonce()));
-		oauth_add_param_to_array(argcp, argvp, oarg);
-		free(tmp);
-	}
+  if (!oauth_param_exists(*argvp,*argcp,"oauth_nonce")) {
+    char *tmp;
+    snprintf(oarg, 1024, "oauth_nonce=%s", (tmp=oauth_gen_nonce()));
+    oauth_add_param_to_array(argcp, argvp, oarg);
+    free(tmp);
+  }
 
-	if (!oauth_param_exists(*argvp,*argcp,"oauth_timestamp")) {
-		snprintf(oarg, 1024, "oauth_timestamp=%li", time(NULL));
-		oauth_add_param_to_array(argcp, argvp, oarg);
-	}
+  if (!oauth_param_exists(*argvp,*argcp,"oauth_timestamp")) {
+    snprintf(oarg, 1024, "oauth_timestamp=%li", (long int) time(NULL));
+    oauth_add_param_to_array(argcp, argvp, oarg);
+  }
 
-	if (t_key) {
+  if (t_key) {
     snprintf(oarg, 1024, "oauth_token=%s", t_key);
-		oauth_add_param_to_array(argcp, argvp, oarg);
+    oauth_add_param_to_array(argcp, argvp, oarg);
   }
 
   snprintf(oarg, 1024, "oauth_consumer_key=%s", c_key);
-	oauth_add_param_to_array(argcp, argvp, oarg);
+  oauth_add_param_to_array(argcp, argvp, oarg);
 
   snprintf(oarg, 1024, "oauth_signature_method=%s",
       method==0?"HMAC-SHA1":method==1?"RSA-SHA1":"PLAINTEXT");
-	oauth_add_param_to_array(argcp, argvp, oarg);
+  oauth_add_param_to_array(argcp, argvp, oarg);
 
-	if (!oauth_param_exists(*argvp,*argcp,"oauth_version")) {
-		snprintf(oarg, 1024, "oauth_version=1.0");
-		oauth_add_param_to_array(argcp, argvp, oarg);
-	}
+  if (!oauth_param_exists(*argvp,*argcp,"oauth_version")) {
+    snprintf(oarg, 1024, "oauth_version=1.0");
+    oauth_add_param_to_array(argcp, argvp, oarg);
+  }
 
 #if 0 // oauth_version 1.0 Rev A
-	if (!oauth_param_exists(argv,argc,"oauth_callback")) {
-		snprintf(oarg, 1024, "oauth_callback=oob");
-		oauth_add_param_to_array(argcp, argvp, oarg);
-	}
+  if (!oauth_param_exists(argv,argc,"oauth_callback")) {
+    snprintf(oarg, 1024, "oauth_callback=oob");
+    oauth_add_param_to_array(argcp, argvp, oarg);
+  }
 #endif
 
 }
 
-/**
- * calculate oAuth-signature for a given request URL, parameters and oauth-tokens.
- *
- * if 'postargs' is NULL a "GET" request is signed and the 
- * signed URL is returned. Else this fn will modify 'postargs' 
- * to point to memory that contains the signed POST-variables 
- * and returns the base URL.
- *
- * both, the return value and (if given) 'postargs' need to be freed
- * by the caller.
- *
- * @param url The request URL to be signed. append all GET or POST 
- * query-parameters separated by either '?' or '&' to this parameter.
- *
- * @param postargs This parameter points to an area where the return value
- * is stored. If 'postargs' is NULL, no value is stored.
- *
- * @param method specify the signature method to use. It is of type 
- * \ref OAuthMethod and most likely \ref OA_HMAC.
- *
- * @param c_key consumer key
- * @param c_secret consumer secret
- * @param t_key token key
- * @param t_secret token secret
- *
- * @return the signed url or NULL if an error occurred.
- *
- */
 char *oauth_sign_url (const char *url, char **postargs, 
   OAuthMethod method, 
   const char *c_key, //< consumer key - posted plain text
@@ -781,6 +756,20 @@ char *oauth_sign_url (const char *url, char **postargs,
   const char *t_key, //< token key - posted plain text in URL
   const char *t_secret //< token secret - used as 2st part of secret-key
   ) {
+  return oauth_sign_url2(url, postargs, 
+    method, NULL,
+    c_key, c_secret,
+    t_key, t_secret);
+} 
+
+char *oauth_sign_url2 (const char *url, char **postargs, 
+  OAuthMethod method, 
+  const char *http_method, //< HTTP request method
+  const char *c_key, //< consumer key - posted plain text
+  const char *c_secret, //< consumer secret - used as 1st part of secret-key 
+  const char *t_key, //< token key - posted plain text in URL
+  const char *t_secret //< token secret - used as 2st part of secret-key
+  ) {
   int  argc;
   char **argv = NULL;
   char *rv;
@@ -790,11 +779,12 @@ char *oauth_sign_url (const char *url, char **postargs,
   else
     argc = oauth_split_url_parameters(url, &argv);
 
-  rv=oauth_sign_array(&argc, &argv, postargs, 
-		method, c_key, c_secret, t_key, t_secret);
+  rv=oauth_sign_array2(&argc, &argv, postargs, 
+    method, http_method,
+    c_key, c_secret, t_key, t_secret);
 
   oauth_free_array(&argc, &argv);
-	return(rv);
+  return(rv);
 }
 
 char *oauth_sign_array (int *argcp, char***argvp,
@@ -805,13 +795,39 @@ char *oauth_sign_array (int *argcp, char***argvp,
   const char *t_key, //< token key - posted plain text in URL
   const char *t_secret //< token secret - used as 2st part of secret-key
   ) {
+  return oauth_sign_array2 (argcp, argvp, 
+                            postargs, method,
+                            NULL,
+                            c_key, c_secret,
+                            t_key, t_secret);
+}
+
+char *oauth_sign_array2 (int *argcp, char***argvp,
+  char **postargs,
+  OAuthMethod method, 
+  const char *http_method, //< HTTP request method
+  const char *c_key, //< consumer key - posted plain text
+  const char *c_secret, //< consumer secret - used as 1st part of secret-key 
+  const char *t_key, //< token key - posted plain text in URL
+  const char *t_secret //< token secret - used as 2st part of secret-key
+  ) {
   char oarg[1024];
   char *query;
   char *okey, *odat, *sign;
   char *result;
+  char *http_request_method;
 
-	// add OAuth protocol parameters
-	oauth_add_protocol(argcp, argvp, method, c_key, t_key);
+  if (!http_method) {
+    http_request_method = xstrdup(postargs?"POST":"GET");
+  } else {
+    int i;
+    http_request_method = xstrdup(http_method);
+    for (i=0;i<strlen(http_request_method);i++) 
+      http_request_method[i]=toupper(http_request_method[i]);
+  }
+
+  // add OAuth protocol parameters
+  oauth_add_protocol(argcp, argvp, method, c_key, t_key);
 
   // sort parameters
   qsort(&(*argvp)[1], (*argcp)-1, sizeof(char *), oauth_cmpstringp);
@@ -821,7 +837,8 @@ char *oauth_sign_array (int *argcp, char***argvp,
 
   // generate signature
   okey = oauth_catenc(2, c_secret, t_secret);
-  odat = oauth_catenc(3, postargs?"POST":"GET", (*argvp)[0], query);
+  odat = oauth_catenc(3, http_request_method, (*argvp)[0], query);
+  free(http_request_method);
 #ifdef DEBUG_OAUTH
   fprintf (stderr, "\nliboauth: data to sign='%s'\n\n", odat);
   fprintf (stderr, "\nliboauth: key='%s'\n\n", okey);
@@ -829,23 +846,23 @@ char *oauth_sign_array (int *argcp, char***argvp,
   switch(method) {
     case OA_RSA:
       sign = oauth_sign_rsa_sha1(odat,okey); // XXX okey needs to be RSA key!
-    	break;
+      break;
     case OA_PLAINTEXT:
       sign = oauth_sign_plaintext(odat,okey);
-    	break;
+      break;
     default:
       sign = oauth_sign_hmac_sha1(odat,okey);
   }
 #ifdef WIPE_MEMORY
-	memset(okey,0, strlen(okey));
-	memset(odat,0, strlen(odat));
+  memset(okey,0, strlen(okey));
+  memset(odat,0, strlen(odat));
 #endif
   free(odat); 
   free(okey);
 
   // append signature to query args.
   snprintf(oarg, 1024, "oauth_signature=%s",sign);
-	oauth_add_param_to_array(argcp, argvp, oarg);
+  oauth_add_param_to_array(argcp, argvp, oarg);
   free(sign);
 
   // build URL params
@@ -869,12 +886,63 @@ char *oauth_sign_array (int *argcp, char***argvp,
  */
 void oauth_free_array(int *argcp, char ***argvp) {
   int i;
-	for (i=0;i<(*argcp);i++) {
-		free((*argvp)[i]);
-	}
+  for (i=0;i<(*argcp);i++) {
+    free((*argvp)[i]);
+  }
   if(*argvp) free(*argvp);
 }
 
+/** 
+ * http://oauth.googlecode.com/svn/spec/ext/body_hash/1.0/drafts/4/spec.html
+ */
+char *oauth_body_hash_file(char *filename) {
+  unsigned char fb[BUFSIZ];
+  EVP_MD_CTX ctx;
+  size_t len=0;
+  unsigned char *md;
+  FILE *F= fopen(filename, "r");
+  if (!F) return NULL;
+
+  EVP_MD_CTX_init(&ctx);
+  EVP_DigestInit(&ctx,EVP_sha1());
+  while (!feof(F) && (len=fread(fb,sizeof(char),BUFSIZ, F))>0) {
+    EVP_DigestUpdate(&ctx, fb, len);
+  }
+  fclose(F);
+  len=0;
+  md=(unsigned char*) calloc(EVP_MD_size(EVP_sha1()),sizeof(unsigned char));
+  EVP_DigestFinal(&ctx, md,(unsigned int*) &len);
+  EVP_MD_CTX_cleanup(&ctx);
+  return oauth_body_hash_encode(len, md);
+}
+
+char *oauth_body_hash_data(size_t length, const char *data) {
+  EVP_MD_CTX ctx;
+  size_t len=0;
+  unsigned char *md;
+  md=(unsigned char*) calloc(EVP_MD_size(EVP_sha1()),sizeof(unsigned char));
+  EVP_MD_CTX_init(&ctx);
+  EVP_DigestInit(&ctx,EVP_sha1());
+  EVP_DigestUpdate(&ctx, data, length);
+  EVP_DigestFinal(&ctx, md,(unsigned int*) &len);
+  EVP_MD_CTX_cleanup(&ctx);
+  return oauth_body_hash_encode(len, md);
+}
+
+/**
+ * base64 encode digest, free it and return a URL parameter
+ * with the oauth_body_hash
+ */
+char *oauth_body_hash_encode(size_t len, unsigned char *digest) {
+  char *sign=oauth_encode_base64(len,digest);
+  char *sig_url = (char*)malloc(17+strlen(sign));
+  sprintf(sig_url,"oauth_body_hash=%s", sign);
+  free(sign);
+  free(digest);
+  return sig_url;
+}
+
+
 /**
  * xep-0235 - TODO
  */
@@ -882,7 +950,7 @@ char *oauth_sign_xmpp (const char *xml,
   OAuthMethod method, 
   const char *c_secret, //< consumer secret - used as 1st part of secret-key 
   const char *t_secret //< token secret - used as 2st part of secret-key
-	) {
+  ) {
 
   return NULL;
 }
