@@ -1,7 +1,7 @@
 /*
  * oAuth string functions in POSIX-C.
  *
- * Copyright 2007, 2008 Robin Gareus <robin@gareus.org>
+ * Copyright 2007, 2008, 2009 Robin Gareus <robin@gareus.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,18 +31,26 @@
 #include <stdlib.h>
 #include <string.h>
 
+#ifdef WIN32
+#  define snprintf _snprintf
+#endif
+
 #include "xmalloc.h"
 #include "oauth.h"
 
 #define OAUTH_USER_AGENT "liboauth-agent/" VERSION
 
-#ifdef HAVE_CURL
+#ifdef HAVE_CURL /* HTTP requests via libcurl */
 #include <curl/curl.h>
 #include <sys/stat.h>
 
 struct MemoryStruct {
   char *data;
-  size_t size;
+  size_t size; //< bytes remaining (r), bytes accumulated (w)
+
+  size_t start_size; //< only used with ..AndCall()
+  void (*callback)(void*,int,size_t,size_t); //< only used with ..AndCall()
+  void *callback_data; //< only used with ..AndCall()
 };
 
 static size_t
@@ -59,6 +67,33 @@ WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data) {
   return realsize;
 }
 
+static size_t
+ReadMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data) {
+  struct MemoryStruct *mem = (struct MemoryStruct *)data;
+  size_t realsize = size * nmemb;
+  if (realsize > mem->size) realsize = mem->size;
+  memcpy(ptr, mem->data, realsize);
+  mem->size -= realsize;
+  mem->data += realsize;
+  return realsize;
+}
+
+static size_t
+WriteMemoryCallbackAndCall(void *ptr, size_t size, size_t nmemb, void *data) {
+  struct MemoryStruct *mem = (struct MemoryStruct *)data;
+  size_t ret=WriteMemoryCallback(ptr,size,nmemb,data);
+  mem->callback(mem->callback_data,0,mem->size,mem->size);
+  return ret;
+}
+
+static size_t
+ReadMemoryCallbackAndCall(void *ptr, size_t size, size_t nmemb, void *data) {
+  struct MemoryStruct *mem = (struct MemoryStruct *)data;
+  size_t ret=ReadMemoryCallback(ptr,size,nmemb,data);
+  mem->callback(mem->callback_data,1,mem->start_size-mem->size,mem->start_size);
+  return ret;
+}
+
 /**
  * cURL http post function.
  * the returned string (if not NULL) needs to be freed by the caller
@@ -106,7 +141,7 @@ char *oauth_curl_get (const char *u, const char *q) {
   struct MemoryStruct chunk;
 
   if (q) {
-    t1=xmalloc(sizeof(char)*(strlen(u)+strlen(q)+2));
+    t1=(char*)xmalloc(sizeof(char)*(strlen(u)+strlen(q)+2));
     strcat(t1,u); strcat(t1,"?"); strcat(t1,q);
   }
 
@@ -118,6 +153,12 @@ char *oauth_curl_get (const char *u, const char *q) {
   curl_easy_setopt(curl, CURLOPT_URL, q?t1:u);
   curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&chunk);
   curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
+#if 0 // TODO - support request methods..
+  if (0) 
+    curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "HEAD");
+  else if (0) 
+    curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "DELETE");
+#endif
   curl_easy_setopt(curl, CURLOPT_USERAGENT, OAUTH_USER_AGENT);
   res = curl_easy_perform(curl);
   if (q) free(t1);
@@ -147,7 +188,7 @@ char *oauth_curl_post_file (const char *u, const char *fn, size_t len, const cha
   FILE *f;
 
   chunk.data=NULL;
-  chunk.size = 0;
+  chunk.size=0;
 
   if (customheader)
     slist = curl_slist_append(slist, customheader);
@@ -184,16 +225,113 @@ char *oauth_curl_post_file (const char *u, const char *fn, size_t len, const cha
   return (chunk.data);
 }
 
-#endif // no cURL.
+/**
+ * http send raw data, with callback.
+ * the returned string needs to be freed by the caller
+ *
+ * more documentation in oauth.h
+ *
+ * @param u url to retrieve
+ * @param data data to post along
+ * @param len length of the file in bytes. set to '0' for autodetection
+ * @param customheader specify custom HTTP header (or NULL for default)
+ * @param callback specify the callback function
+ * @param callback_data specify data to pass to the callback function
+ * @return returned HTTP reply or NULL on error
+ */
+char *oauth_curl_send_data_with_callback (const char *u, const char *data, size_t len, const char *customheader, void (*callback)(void*,int,size_t,size_t), void *callback_data, const char *httpMethod) {
+  CURL *curl;
+  CURLcode res;
+  struct curl_slist *slist=NULL;
+  struct MemoryStruct chunk;
+  struct MemoryStruct rdnfo;
+  
+  chunk.data=NULL;
+  chunk.size=0;
+  chunk.start_size=0;
+  chunk.callback=callback;
+  chunk.callback_data=callback_data;
+  rdnfo.data=(char *)data;
+  rdnfo.size=len;
+  rdnfo.start_size=len;
+  rdnfo.callback=callback;
+  rdnfo.callback_data=callback_data;
+
+  if (customheader)
+    slist = curl_slist_append(slist, customheader);
+  else
+    slist = curl_slist_append(slist, "Content-Type: image/jpeg;");
+  
+  curl = curl_easy_init();
+  if(!curl) return NULL;
+  curl_easy_setopt(curl, CURLOPT_URL, u);
+  curl_easy_setopt(curl, CURLOPT_POST, 1);
+  if (httpMethod) curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, httpMethod);
+  curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, len);
+  curl_easy_setopt(curl, CURLOPT_HTTPHEADER, slist); 
+  curl_easy_setopt(curl, CURLOPT_READDATA, (void *)&rdnfo);
+  if (callback)
+    curl_easy_setopt(curl, CURLOPT_READFUNCTION, ReadMemoryCallbackAndCall);
+  else
+    curl_easy_setopt(curl, CURLOPT_READFUNCTION, ReadMemoryCallback);
+  curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&chunk);
+  if (callback)
+  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallbackAndCall);
+  else 
+     curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
+  curl_easy_setopt(curl, CURLOPT_USERAGENT, OAUTH_USER_AGENT);
+  res = curl_easy_perform(curl);
+  if (res) {
+    // error
+    return NULL;
+  }
+  
+  curl_easy_cleanup(curl);
+  return (chunk.data);
+}
+
+/**
+ * http post raw data.
+ * the returned string needs to be freed by the caller
+ *
+ * more documentation in oauth.h
+ *
+ * @param u url to retrieve
+ * @param data data to post along
+ * @param len length of the file in bytes. set to '0' for autodetection
+ * @param customheader specify custom HTTP header (or NULL for default)
+ * @return returned HTTP reply or NULL on error
+ */
+char *oauth_curl_post_data(const char *u, const char *data, size_t len, const char *customheader) {
+  return oauth_curl_send_data_with_callback(u, data, len, customheader, NULL, NULL, NULL);
+}
+
+char *oauth_curl_send_data (const char *u, const char *data, size_t len, const char *customheader, const char *httpMethod) {
+  return oauth_curl_send_data_with_callback(u, data, len, customheader, NULL, NULL, httpMethod);
+}
+
+char *oauth_curl_post_data_with_callback (const char *u, const char *data, size_t len, const char *customheader, void (*callback)(void*,int,size_t,size_t), void *callback_data) {
+  return oauth_curl_send_data_with_callback(u, data, len, customheader, callback, callback_data, NULL);
+}
+
+#endif // libcURL.
+
+
+#ifdef HAVE_SHELL_CURL /* HTTP requests via command-line curl  */
 
 // command line presets and ENV variable name
 #define _OAUTH_ENV_HTTPCMD "OAUTH_HTTP_CMD"
-#define _OAUTH_DEF_HTTPCMD "curl -sA '"OAUTH_USER_AGENT"' -d '%p' '%u' "
-// alternative: "wget -q -U 'liboauth-agent/0.1' --post-data='%p' '%u' "
-
 #define _OAUTH_ENV_HTTPGET "OAUTH_HTTP_GET_CMD"
-#define _OAUTH_DEF_HTTPGET "curl -sA '"OAUTH_USER_AGENT"' '%u' "
-// alternative: "wget -q -U 'liboauth-agent/0.1' '%u' "
+
+#ifndef _OAUTH_DEF_HTTPCMD
+# define _OAUTH_DEF_HTTPCMD "curl -sA '"OAUTH_USER_AGENT"' -d '%p' '%u' "
+//alternative: "wget -q -U 'liboauth-agent/0.1' --post-data='%p' '%u' "
+#endif
+
+#ifndef _OAUTH_DEF_HTTPGET 
+# define _OAUTH_DEF_HTTPGET "curl -sA '"OAUTH_USER_AGENT"' '%u' "
+//alternative: "wget -q -U 'liboauth-agent/0.1' '%u' "
+#endif
 
 #include <stdio.h>
 
@@ -227,7 +365,7 @@ char *oauth_escape_shell (const char *cmd) {
   int idx;
   while ((tmp=strchr(tmp,'\''))) { 
     idx = tmp-esc;
-    esc=xrealloc(esc,(strlen(esc)+5)*sizeof(char));
+    esc=(char*)xrealloc(esc,(strlen(esc)+5)*sizeof(char));
     memmove(esc+idx+4,esc+idx+1, strlen(esc+idx));
     esc[idx+1]='\\'; esc[idx+2]='\''; esc[idx+3]='\'';
     tmp=esc+(idx+4);
@@ -257,7 +395,7 @@ char *oauth_exec_shell (const char *cmd) {
   int rcv = 1;
   while (in && rcv > 0 && !feof(in)) {
     alloc +=1024;
-    data = xrealloc(data, alloc * sizeof(char));
+    data = (char*)xrealloc(data, alloc * sizeof(char));
     rcv = fread(data, sizeof(char), 1024, in);
     len += rcv;
   }
@@ -295,8 +433,8 @@ char *oauth_exec_post (const char *u, const char *p) {
   t1=strstr(cmdtpl, "%p");
   t2=strstr(cmdtpl, "%u");
   if (!t1 || !t2) {
-	fprintf(stderr, "\nliboauth: invalid HTTP command. set the '%s' environement variable.\n\n",_OAUTH_ENV_HTTPCMD);
-	return(NULL);
+    fprintf(stderr, "\nliboauth: invalid HTTP command. set the '%s' environement variable.\n\n",_OAUTH_ENV_HTTPCMD);
+    return(NULL);
   }
   // TODO: check if there are exactly two '%' in cmdtpl
   *(++t1)= 's'; *(++t2)= 's';
@@ -340,8 +478,8 @@ char *oauth_exec_get (const char *u, const char *q) {
   // add URL and post param - error if no '%p' or '%u' present in definition
   t1=strstr(cmdtpl, "%u");
   if (!t1) {
-	fprintf(stderr, "\nliboauth: invalid HTTP command. set the '%s' environement variable.\n\n",_OAUTH_ENV_HTTPGET);
-	return(NULL);
+    fprintf(stderr, "\nliboauth: invalid HTTP command. set the '%s' environement variable.\n\n",_OAUTH_ENV_HTTPGET);
+    return(NULL);
   }
   *(++t1)= 's';
 
@@ -349,7 +487,7 @@ char *oauth_exec_get (const char *u, const char *q) {
   if (q) {
     char *e2;
     e2 = oauth_escape_shell(q);
-    t1=xmalloc(sizeof(char)*(strlen(e1)+strlen(e2)+2));
+    t1=(char*)xmalloc(sizeof(char)*(strlen(e1)+strlen(e2)+2));
     strcat(t1,e1); strcat(t1,"?"); strcat(t1,e2);
     free(e2);
   }
@@ -359,6 +497,9 @@ char *oauth_exec_get (const char *u, const char *q) {
   if (q) free(t1);
   return oauth_exec_shell(cmd);
 }
+#endif // command-line curl.
+
+/* wrapper functions */
 
 /**
  * do a HTTP GET request, wait for it to finish 
@@ -375,8 +516,10 @@ char *oauth_exec_get (const char *u, const char *q) {
 char *oauth_http_get (const char *u, const char *q) {
 #ifdef HAVE_CURL
   return oauth_curl_get(u,q);
-#else // no cURL.
+#elif defined(HAVE_SHELL_CURL)
   return oauth_exec_get(u,q);
+#else 
+  return NULL;
 #endif
 }
 
@@ -395,8 +538,10 @@ char *oauth_http_get (const char *u, const char *q) {
 char *oauth_http_post (const char *u, const char *p) {
 #ifdef HAVE_CURL
   return oauth_curl_post(u,p);
-#else // no cURL.
+#elif defined(HAVE_SHELL_CURL)
   return oauth_exec_post(u,p);
+#else 
+  return NULL;
 #endif
 }
 
@@ -412,11 +557,58 @@ char *oauth_http_post (const char *u, const char *p) {
  * @param customheader specify custom HTTP header (or NULL for default)
  * @return returned HTTP reply or NULL on error
  */
-char *oauth_post_file (const char *u, const char *fn, const size_t len, const char *contenttype){
+char *oauth_post_file (const char *u, const char *fn, const size_t len, const char *customheader){
 #ifdef HAVE_CURL
-  return oauth_curl_post_file (u, fn, len, contenttype);
-#else
+  return oauth_curl_post_file (u, fn, len, customheader);
+#elif defined(HAVE_SHELL_CURL)
   fprintf(stderr, "\nliboauth: oauth_post_file requires libcurl. libcurl is not available.\n\n");
+  return NULL;
+#else 
+  return NULL;
+#endif
+}
+
+/**
+ * http post raw data.
+ * the returned string needs to be freed by the caller
+ *
+ * more documentation in oauth.h
+ *
+ * @param u url to retrieve
+ * @param data data to post along
+ * @param len length of the file in bytes. set to '0' for autodetection
+ * @param customheader specify custom HTTP header (or NULL for default)
+ * @return returned HTTP reply or NULL on error
+ */
+char *oauth_post_data (const char *u, const char *data, size_t len, const char *customheader) {
+#ifdef HAVE_CURL
+  return oauth_curl_post_data (u, data, len, customheader);
+#elif defined(HAVE_SHELL_CURL)
+  fprintf(stderr, "\nliboauth: oauth_post_file requires libcurl. libcurl is not available.\n\n");
+  return NULL;
+#else
+  return (NULL);
+#endif
+}
+
+char *oauth_send_data (const char *u, const char *data, size_t len, const char *customheader, const char *httpMethod) {
+#ifdef HAVE_CURL
+  return oauth_curl_send_data (u, data, len, customheader, httpMethod);
+#elif defined(HAVE_SHELL_CURL)
+  fprintf(stderr, "\nliboauth: oauth_send_file requires libcurl. libcurl is not available.\n\n");
+  return NULL;
+#else
+  return (NULL);
+#endif
+}
+
+char *oauth_post_data_with_callback (const char *u, const char *data, size_t len, const char *customheader, void (*callback)(void*,int,size_t,size_t), void *callback_data) {
+#ifdef HAVE_CURL
+  return oauth_curl_post_data_with_callback(u, data, len, customheader, callback, callback_data);
+#elif defined(HAVE_SHELL_CURL)
+  fprintf(stderr, "\nliboauth: oauth_post_data_with_callback requires libcurl.\n\n");
+  return NULL;
+#else
   return (NULL);
 #endif
 }
